import { ResourceType, VariableInfo, WgslReflect } from 'wgsl_reflect'
import { WebGPUObjectError } from '../utils/errors.js'
import { BindingMap, GroupBindingMap, VariableStageInfo } from '../utils/bindings.js'

import {
  accessToBufferType,
  accessToStorageTextureAccess,
  parseTextureType,
  typeToSamplerBindingType,
  typeToTextureSampleType,
  typeToViewDimension,
  wgslToWgpuFormat
} from '../utils/wgsl-to-wgpu.js'
import { BufferBindingType } from '../enum.js'

/** @import { WGSLAccess, WGSLSamplerType } from '../utils/wgsl-to-wgpu.js' */

export class ShaderModule {
  _handle
  _code

  /** @type {WgslReflect | undefined} */
  _reflection

  get handle() {
    return this._handle
  }

  get label() {
    return this._handle.label
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUShaderModuleDescriptor} descriptor
   */
  constructor(device, descriptor) {
    this._code = descriptor.code

    try {
      this._handle = device.createShaderModule(descriptor)
    } catch (err) {
      throw WebGPUObjectError.from(err, ShaderModule)
    }
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUShaderModuleDescriptor} descriptor
   */
  static create(device, descriptor) {
    return new ShaderModule(device, descriptor)
  }

  reflect() {
    if (this._reflection == null) {
      this._reflection = new WgslReflect(this._code)
      // no longer needed allow the GC to collect it
      this._code = undefined
    }

    return this._reflection
  }
}

/**
 * @typedef FragmentStateDescriptor
 * @property {Record<string, GPUPipelineConstantValue>} [constants={}]
 * @property {GPUColorTargetState[]} [targets=[]]
 *
 * @typedef VertexStateDescriptor
 * @property {Record<string, GPUPipelineConstantValue>} [constants={}]
 * @property {GPUVertexBufferLayout[]} [buffers=[]]
 *
 * @typedef ShaderPairStateDescriptor
 * @property {FragmentStateDescriptor} [fragment]
 * @property {VertexStateDescriptor} vertex
 *
 */

export class ReflectedShader {
  static _reflectTypes = ['uniforms', 'storage', 'textures', 'samplers']
  _module

  get module() {
    return this._module
  }


  /**
   * @param {ShaderModule} shader
   */
  constructor(shader) {
    this._module = shader
  }


  /**
   * @param {string} name
   * @param {number} group 
   * @returns {VariableInfo | undefined}
   */
  findVariableInfo(name, group) {
    const reflection = this.module.reflect()
    for (const type of ReflectedShader._reflectTypes) {
      const variables = reflection[type]
      if (variables) {
        const variable = variables.find(v => v.name === name && v.group === group)
        if (variable) {
          return variable
        }
      }
    }
  }

  /**
   * @param {string} stageName
   * @returns {string | undefined}
   */
  getEntrypoint(stageName) {
    const entry = this.module.reflect().entry

    // TODO: determine how to correctly handle
    // multiple entrypoints to the same stage
    return entry[stageName].length === 1 ?
      entry[stageName][0].name : undefined
  }

  /**
   * @returns {GPUShaderStageFlags}
   */
  getShaderStages() {
    const entry = this._module.reflect().entry
    let stages = 0

    stages |= entry.vertex.length > 0 ?
      GPUShaderStage.VERTEX : 0

    stages |= entry.fragment.length > 0 ?
      GPUShaderStage.FRAGMENT : 0

    stages |= entry.compute.length > 0 ?
      GPUShaderStage.COMPUTE : 0

    return stages
  }

  /**
   * @param {GPUShaderStageFlags} stages
   */
  hasStage(stages) {
    return this.getShaderStages()
      & stages
  }

  /**
   * @param {GPUShaderStageFlags} stages
   * @param {GroupBindingMap} [out=new GroupBindingMap()]
   */
  getBindingsForStage(stages, out = new GroupBindingMap()) {
    const groups = this._module.reflect().getBindGroups()

    groups.forEach((bindings, groupIndex) => {
      if (!out.has(groupIndex)) {
        out.set(groupIndex, new BindingMap())
      }

      const bindingsMap = out.get(groupIndex)

      bindings.filter(x => x != null).forEach(variableInfo => {
        if (!bindingsMap.has(variableInfo.binding)) {
          bindingsMap.set(variableInfo.binding, { variableInfo, stages })
        } else {
          bindingsMap.get(variableInfo.binding).stages |= stages
        }
      })
    })

    return out
  }

  /**
   * @param {Map<any, any>} map 
   * @returns {number[]}
   */
  static _sortKeyIndices(map) {
    return Array.from(map.keys()).sort((a, b) => a - b)
  }

  /**
   * @param {VariableInfo} _variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  static _parseUniform(_variableInfo) {
    return {
      type: BufferBindingType.Uniform,
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUBufferBindingLayout}
   */
  static _parseStorage(variableInfo) {
    return {
      type: accessToBufferType(
        /** @type {WGSLAccess} */
        (variableInfo.access)
      ),
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  /**
   * @param {VariableInfo} variableInfo
   * @returns {GPUTextureBindingLayout}
   */
  static _parseTexture(variableInfo) {
    const [type, sampledType] = parseTextureType(
      variableInfo.type.name
    )

    return {
      sampleType: typeToTextureSampleType(type, sampledType),
      viewDimension: typeToViewDimension(type),
      multisampled: type.includes('multisampled')
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUSamplerBindingLayout}
   */
  static _parseSampler(variableInfo) {
    return {
      type: typeToSamplerBindingType(
        /** @type {WGSLSamplerType} */(variableInfo.type.name)
      )
    }
  }

  /**
   * @param {VariableInfo} variableInfo 
   * @returns {GPUStorageTextureBindingLayout}
   */
  static _parseStorageTexture(variableInfo) {
    const [type] = parseTextureType(variableInfo.type.name)

    return {
      access: accessToStorageTextureAccess(
        /** @type {WGSLAccess} */(variableInfo.access)
      ),
      format: wgslToWgpuFormat(variableInfo.type.name),
      viewDimension: typeToViewDimension(type)
    }
  }

  /**
   * @param {VariableStageInfo} variableStageInfo
   * @returns {GPUBindGroupLayoutEntry}
   */
  static _variableInfoToEntry(variableStageInfo) {
    const { stages: visibility, variableInfo } = variableStageInfo

    switch (variableInfo.resourceType) {
      case ResourceType.Uniform:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseUniform(variableInfo)
        }
      case ResourceType.Storage:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseStorage(variableInfo)
        }
      case ResourceType.Texture:
        return {
          binding: variableInfo.binding,
          visibility,
          texture: this._parseTexture(variableInfo)
        }
      case ResourceType.Sampler:
        return {
          binding: variableInfo.binding,
          visibility,
          sampler: this._parseSampler(variableInfo)
        }
      case ResourceType.StorageTexture:
        return {
          binding: variableInfo.binding,
          visibility,
          storageTexture: this._parseStorageTexture(variableInfo)
        }
      default:
        console.warn(`Unsupported resource type for reflection: ${ResourceType[variableInfo.resourceType]}`)
        return
    }
  }

  /**
   * @param {GroupBindingMap} groupBindings
   */
  static createBindGroupLayoutEntries(groupBindings) {
    const sortedGroupIndices = this._sortKeyIndices(groupBindings)

    return sortedGroupIndices.map(groupIndex => {
      const bindingsMap = groupBindings.get(groupIndex)
      return this._sortKeyIndices(bindingsMap).map(bindingIndex => (
        this._variableInfoToEntry(bindingsMap.get(bindingIndex))
      ))
    })
      .filter(desc => desc.length > 0)
  }
}

export class ShaderPair {
  /** @type {ReflectedShader} */
  _vertex
  /** @type {ReflectedShader} */
  _fragment

  /**
   * @param {ReflectedShader} vertex
   * @param {ReflectedShader} [fragment]
   */
  constructor(vertex, fragment) {
    if (!vertex) {
      throw new Error('Missing vertex shader')
    }

    if (!vertex.hasStage(GPUShaderStage.VERTEX)) {
      throw new Error('Vertex shader does not have a vertex entrypoint.')
    }

    this._vertex = vertex

    if (fragment) {
      if (fragment.hasStage(GPUShaderStage.FRAGMENT)) {
        this._fragment = fragment
      } else {
        throw new Error('Fragment shader does not have a fragment entrypoint.')
      }
    } else if (this._vertex.hasStage(GPUShaderStage.FRAGMENT)) {
      // this is a unified shader, use the vertex as the fragment
      this._fragment = vertex
    } else {
      throw new Error('Missing fragment shader.')
    }
  }

  /** @param {ShaderModule} shader */
  static fromUnifiedShader(shader) {
    return new ShaderPair(
      new ReflectedShader(shader)
    )
  }

  /**
   * @param {{
       vertex: ShaderModule,
       fragment?: ShaderModule
   * }} value
   */
  static fromPair(value) {
    const vert = new ReflectedShader(value.vertex)
    const frag = value.fragment && new ReflectedShader(value.fragment)
    return new ShaderPair(vert, frag)
  }

  _createGroupBindings() {
    const groupBindings = new GroupBindingMap()

    this._vertex.getBindingsForStage(
      GPUShaderStage.VERTEX,
      groupBindings
    )

    this._fragment.getBindingsForStage(
      GPUShaderStage.FRAGMENT,
      groupBindings
    )

    return groupBindings
  }

  createBindGroupLayoutEntries() {
    return ReflectedShader.createBindGroupLayoutEntries(
      this._createGroupBindings()
    )
  }

  /**
   * @param {FragmentStateDescriptor} descriptor
   * @returns {GPUFragmentState}
   */
  _getFragmentState(descriptor) {
    return {
      module: this._fragment.module.handle,
      entryPoint: this._fragment.getEntrypoint('fragment'),
      constants: descriptor.constants || {},
      targets: descriptor.targets || []
    }
  }

  /**
   * @param {VertexStateDescriptor} descriptor
   * @returns {GPUVertexState}
   */
  _getVertexState(descriptor) {
    return {
      module: this._vertex.module.handle,
      entryPoint: this._vertex.getEntrypoint('vertex'),
      constants: descriptor.constants || {},
      buffers: descriptor.buffers || []
    }
  }

  /**
   * @param {string} name
   * @param {number} group 
   * @returns {VariableInfo | undefined}
   */
  findVariableInfo(name, group) {
    let variableInfo = this._vertex.findVariableInfo(name, group)

    if (!variableInfo && this._fragment !== this._vertex) {
      variableInfo = this._fragment.findVariableInfo(name, group)
    }

    return variableInfo
  }

  /**
   * @param {ShaderPairStateDescriptor} descriptor
   * @returns {Pick<GPURenderPipelineDescriptor, 'vertex' | 'fragment'>}
   */
  getRenderPipelineStates(descriptor) {
    return {
      fragment: this._getFragmentState(descriptor.fragment),
      vertex: this._getVertexState(descriptor.vertex),
    }
  }
}

